// SPDX-FileCopyrightText: 2021-2025 The Ikarus Developers ikarus@ibb.uni-stuttgart.de
// SPDX-License-Identifier: LGPL-3.0-or-later

/**
 * \file asfunctions.hh
 * \brief Header file for various assumed stress functions.
 * \ingroup ps
 */

#pragma once

#include <ikarus/finiteelements/mechanics/assumedstress/asfunctions/linearstress.hh>
#include <ikarus/finiteelements/mechanics/assumedstress/asfunctions/pk2stress.hh>
